<?php

// autoload_files.php @generated by Composer

$vendorDir = dirname(dirname(__FILE__));
$baseDir = dirname($vendorDir);

return array(
    '7b11c4dc42b3b3023073cb14e519683c' => $vendorDir . '/ralouphie/getallheaders/src/getallheaders.php',
    'e3e111437f37e10e6bcab5eacc08fb6f' => $baseDir . '/vendor_prefixed/guzzlehttp/promises/src/functions_include.php',
    '2bb094e40611cb5eccea789f32aff634' => $baseDir . '/vendor_prefixed/symfony/polyfill-mbstring/bootstrap.php',
    '1fd84176824b5a44e7bd8da85eca7e14' => $baseDir . '/vendor_prefixed/symfony/polyfill-php72/bootstrap.php',
    '606299e0d90ec13f1e6b53164b8387df' => $baseDir . '/vendor_prefixed/symfony/polyfill-intl-idn/bootstrap.php',
    '2d822e735b5b1897d96a7a28221d6513' => $baseDir . '/vendor_prefixed/symfony/deprecation-contracts/function.php',
    '6fe0d6ea1deb6acc74bbe64573a83e1c' => $baseDir . '/vendor_prefixed/guzzlehttp/guzzle/src/functions_include.php',
    '3ed0dcebed83aa26dfe4c549d730cf2e' => $baseDir . '/vendor_prefixed/phpseclib/phpseclib/phpseclib/bootstrap.php',
);
